use std::{
    error,
    fmt,
    borrow::Cow,
};

pub type FrozenVec<T> = Box<[T]>;
pub type FrozenString = Box<str>;

// NOTE: Currently unused, as we use `to_utf8_lossy()` for PCRE2 `byte`-matching (XXX: Should we change?)
// TODO: to return some kind of `Either<&'s str, impl bytes::Buf + 's>` type, which would use `str` on non-PCRE, but opaque `bytes::Buf` on PCRE?)
pub type FrozenBytes = FrozenVec<u8>; 

pub type Groups<String = FrozenString> = FrozenVec<Option<String>>; // TODO: See `exec()` comment below about named groups and switching to `BTreeMap<{enum : CaptureGroupIdent::Index, Name}, Option<Cow<'s str>>>`

#[derive(Debug, Clone)]
pub struct Regex
{
    #[cfg(feature="perl")]
    internal: pcre2::bytes::Regex,
    #[cfg(not(feature = "perl"))]
    internal: regex::Regex,
}

#[derive(Debug)]
pub enum Error
{
    Compile(String),
    Execute,
    Internal,
    Unknown,
}

impl error::Error for Error{}
impl fmt::Display for Error
{
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
    {
	write!(f, "regex error: ")?;
	match self {
	    Error::Compile(s) => write!(f, "compilation failed: {}", s),
	    Error::Execute => write!(f, "execution failed"),
	    Error::Internal => write!(f, "internal"),
	    _ => write!(f, "unknown"),
	}
    }
}

impl Regex {
    pub fn compile(string: impl AsRef<str>) -> Result<Self, Error> 
    {
	#[cfg(feature = "perl")]
	return Ok(Self{internal: pcre2::bytes::RegexBuilder::new().build(string.as_ref())?});
	#[cfg(not(feature = "perl"))]
	return Ok(Self{internal: regex::Regex::new(string.as_ref())?});
    }

    pub fn exec<'s>(&self, string: &'s str) -> Result<Option<Groups<Cow<'s, str>>>, Error> //TODO: Can we also add named groups with a `BTreeMap<{CG::Index(usize) | CG:Name(String)}, Option<Groups<Cow<'s, str>>>>` (XXX: And maybe also be able to simplefy `V` to just `Option<&'s str>` / `Option<Cow<'s, str>>`, since the group index is already encoded in `K` (group index / group name mapped to potential match of associated group).)
    {
	#[cfg(feature = "perl")]
	return {
	    Ok(match self.internal.captures(string.as_ref())? {
		Some(m) => {
		    Some((0..m.len()).map(move |i| m.get(i).map(|x| String::from_utf8_lossy(x.as_bytes()) )).collect())
		},
		None => None,
	    })
	};
	#[cfg(not(feature = "perl"))]
	return {
	    Ok(match self.internal.captures(string.as_ref()) {
		Some(m) => {
		    Some((0..m.len()).map(move |i| m.get(i).map(|x| Cow::Borrowed(x.as_str()) )).collect())
		},
		None => None,
	    })
	};
    }
}

impl From<fmt::Error> for Error
{
    fn from(_er: fmt::Error) -> Self
    {
	Self::Internal
    }
}

//#[cfg(not(feature = "perl"))]
impl From<regex::Error> for Error
{
    fn from(er: regex::Error) -> Self
    {
	Self::Compile(format!("{}", er))
    }
}

#[cfg(feature = "perl")]
impl From<pcre2::Error> for Error
{
    fn from(er: pcre2::Error) -> Self
    {
	Self::Compile(format!("{}", er))
    }
}
